import datetime
import zoneinfo

import pytest

from jsno import jsonify, unjsonify, UnjsonifyError


helsinki = zoneinfo.ZoneInfo("Europe/Helsinki")
ouagadougou = zoneinfo.ZoneInfo("Africa/Ouagadougou")
utc = zoneinfo.ZoneInfo("UTC")


def test_jsonify_dates():

    assert (
        jsonify(
            [
                datetime.datetime(2023, 7, 15, 8, 39, 0, tzinfo=helsinki),
                datetime.datetime(2023, 7, 15, 8, 40, 10, tzinfo=utc),
                datetime.datetime(2023, 7, 15, 8, 41, 20),
                datetime.datetime(2023, 7, 15, 8, 39, 0, 1, tzinfo=helsinki),
                datetime.datetime(2023, 7, 15, 8, 40, 10, 120, tzinfo=utc),
                datetime.datetime(2023, 7, 15, 8, 40, 10, 1, tzinfo=datetime.timezone.utc),
                datetime.datetime(5012, 7, 15, 8, 41, 20, 123400),
                datetime.datetime(2023, 7, 29, 13, 23, 11, tzinfo=ouagadougou)
            ]
        )
        ==
        [
            "2023-07-15T08:39:00+03:00",
            "2023-07-15T08:40:10Z",
            "2023-07-15T08:41:20",
            "2023-07-15T08:39:00.000001+03:00",
            "2023-07-15T08:40:10.000120Z",
            "2023-07-15T08:40:10.000001Z",
            "5012-07-15T08:41:20.123400",
            "2023-07-29T13:23:11Z",
        ]
    )


def test_jsonify_old_dates():
    assert (
        jsonify(
            [
                datetime.datetime(1940, 7, 15, 8, 39, 0, tzinfo=helsinki),
                datetime.datetime(1960, 7, 15, 8, 39, 0, tzinfo=helsinki),
                datetime.datetime(1980, 7, 15, 8, 39, 0, tzinfo=helsinki),
                datetime.datetime(1981, 7, 15, 8, 39, 0, tzinfo=helsinki),
            ]
        )
        ==
        [
            "1940-07-15T08:39:00+02:00",
            "1960-07-15T08:39:00+02:00",
            "1980-07-15T08:39:00+02:00",
            "1981-07-15T08:39:00+03:00",
        ]
    )


def test_unjsonify_datetimes():
    assert (
        unjsonify[list[datetime.datetime]](
            [
                "2023-07-15T08:39:00+03:00",
                "2023-07-15T08:40:10Z",
                "2023-07-15T08:41:20",
                "2023-07-15T08:39:00.000001+03:00",
                "2023-07-15T08:40:10.000120Z",
                "5012-07-15T08:41:20.123400",
            ]
        ) == (
            [
                datetime.datetime(2023, 7, 15, 8, 39, 0, tzinfo=helsinki),
                datetime.datetime(2023, 7, 15, 8, 40, 10, tzinfo=utc),
                datetime.datetime(2023, 7, 15, 8, 41, 20),
                datetime.datetime(2023, 7, 15, 8, 39, 0, 1, tzinfo=helsinki),
                datetime.datetime(2023, 7, 15, 8, 40, 10, 120, tzinfo=utc),
                datetime.datetime(5012, 7, 15, 8, 41, 20, 123400),
            ]
        )
    )


def test_unjsonify_datetime_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[datetime.datetime]('2023-13-13T12:34:56')


def test_jsonify_time():
    assert jsonify(datetime.time(23, 59, 59)) == "23:59:59"


def test_unjsonify_time():
    assert unjsonify[datetime.time]("23:59:59") == datetime.time(23, 59, 59)


def test_unjsonify_time_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[datetime.time]("24:59:59")


def test_unjsonify_date_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[datetime.date]("today")


def test_jsonify_timedelta_days_only():
    assert jsonify(datetime.timedelta(days=23)) == "23 days, 0:00:00"


def test_jsonify_timedelta_microsecond():
    assert jsonify(datetime.timedelta(microseconds=1)) == "0:00:00.000001"


def test_jsonify_timedelta_negative_days():
    assert jsonify(datetime.timedelta(days=-123)) == "-123 days, 0:00:00"


def test_unjsonify_timedelta_days_only():
    assert unjsonify[datetime.timedelta]("23 days, 0:00:00") == datetime.timedelta(days=23)


def test_unjsonify_timedelta_negative():
    assert (
        unjsonify[datetime.timedelta]("-10 days, 1:23:45") ==
        datetime.timedelta(days=-11, hours=22, minutes=36, seconds=15)
    )


def test_unjsonify_timedelta_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[datetime.timedelta]("-10 days, 24:23:45")


def test_unjsonify_timezone_utc():
    assert unjsonify[datetime.timezone]("UTC") == datetime.timezone.utc


def test_unjsonify_timezone_utc_positive():
    assert (
        unjsonify[datetime.timezone]("UTC+01:00:00") ==
        datetime.timezone(datetime.timedelta(hours=1))
    )


def test_unjsonify_timezone_utc_negative():
    assert (
        unjsonify[datetime.timezone]("UTC-03:30:00") ==
        datetime.timezone(datetime.timedelta(hours=-3, minutes=-30))
    )


def test_unjsonify_timezone_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[datetime.timezone]("UTC03:30:00")



"""
Jsonification and unjsonification for standard Python types.

* datetime.datetime
* datetime.date
* datetime.time
* datetime.timezone
* decimal.Decimal
* pathlib.Path
* zoneinfo.ZoneInfo
* enum.Enum
* ranges

"""

import dataclasses
import decimal
import enum
import pathlib
import re
import uuid
import zoneinfo

from types import NoneType, SimpleNamespace
from typing import Any

from jsno.jsonify import jsonify, JSON
from jsno.unjsonify import unjsonify, typecheck, raise_error, cast


# marking types to be jsonified as strings


def jsonify_to_string(value: Any) -> str:
    return str(value)


def unjsonify_from_string(value: JSON, as_type: type) -> Any:
    typecheck(value, str, as_type)
    return cast(value, as_type)


def jsonify_as_string(type_: type, exceptions: type | tuple[type, ...] = ()) -> type:
    jsonify.register(type_)(jsonify_to_string)

    if exceptions:

        @unjsonify.register(type_)
        def _(value, as_type):
            typecheck(value, str, as_type)
            try:
                return cast(value, as_type)
            except Exception as exc:
                if not isinstance(exc, exceptions):
                    raise

                detail = exc.args[0]

            raise_error(value, as_type, detail)

    else:
        unjsonify.register(type_)(unjsonify_from_string)

    return type_


# NoneType, jsonify is not needed

@unjsonify.register(NoneType)
def _(value, as_type):
    typecheck(value, NoneType, as_type)
    return None


# str

jsonify_as_string(str)

# bool, jsonify is not needed


@unjsonify.register(bool)
def _(value, as_type):
    typecheck(value, bool, as_type)
    return cast(value, as_type)


# int


@jsonify.register(int)
def _(value):
    return int(value)


@unjsonify.register(int)
def _(value, as_type):
    typecheck(value, int, as_type)
    return cast(value, as_type)


# float


@jsonify.register(float)
def _(value):
    return float(value)


@unjsonify.register(float)
def _(value, as_type):
    typecheck(value, (float, int), as_type)
    return cast(value, as_type)


# enums


@jsonify.register(enum.Enum)
def _(enum):
    return enum.name


@unjsonify.register(enum.Enum)
def _(value, as_type):
    typecheck(value, str, as_type)
    try:
        return getattr(as_type, value)
    except AttributeError:
        pass

    raise_error(value, as_type)


# zoneinfo

jsonify_as_string(zoneinfo.ZoneInfo, exceptions=(zoneinfo.ZoneInfoNotFoundError))

# decimal


@jsonify.register(decimal.Decimal)
def _(value):
    """
    Always jsnonify Decimal as a string.
    """
    return str(value)


@unjsonify.register(decimal.Decimal)
def _(value, as_type):
    typecheck(value, (str, int), as_type)

    return as_type(value)


# pathlib.Path

jsonify_as_string(pathlib.Path)


# uuid.UUID

jsonify_as_string(uuid.UUID)


# complex numbers

jsonify_as_string(complex)


@jsonify.register(range)
def _(value):
    result = {"start": value.start, "stop": value.stop}
    if value.step != 1:
        result["step"] = value.step
    return result


@dataclasses.dataclass(slots=True, frozen=True)
class Range:
    start: int
    stop: int
    step: int | None = None


@unjsonify.register(range)
def _(value, as_type):
    it = unjsonify[Range](value)
    if it.step == 0:
        raise_error(value, as_type, "Range step must not be zero")

    return as_type(it.start, it.stop, it.step or 1)


# re.Pattern


@jsonify.register(re.Pattern)
def _(value):
    return value.pattern


@unjsonify.register(re.Pattern)
def _(value, as_type):
    return re.compile(value)


# types.SimpleNamespace

@jsonify.register(SimpleNamespace)
def _(value):
    return jsonify(value.__dict__)


@unjsonify.register(SimpleNamespace)
def _(value, as_type):
    typecheck(value, dict, as_type)
    return as_type(**value)



from datetime import datetime
from typing import TypedDict, NotRequired, Required, NamedTuple, Optional

import pytest

from jsno import jsonify, unjsonify, UnjsonifyError, property_name, Constraint


class ApiKey(TypedDict):
    name: NotRequired[str]
    value: str
    created_at: datetime


def test_unjsonify_typed_dict():
    apikey = unjsonify[ApiKey]({"value": "XXX", "created_at": "2023-08-05T07:22:33"})
    assert apikey == {"value": "XXX", "created_at": datetime(2023, 8, 5, 7, 22, 33)}


class ApiKey2(TypedDict, total=False):
    name: str
    value: Required[str]
    created_at: datetime


def test_unjsonify_nontotal_typed_dict():
    apikey = unjsonify[ApiKey2]({"value": "XXX", "created_at": "2023-08-05T07:22:33"})
    assert apikey == {"value": "XXX", "created_at": datetime(2023, 8, 5, 7, 22, 33)}


def test_unjsonify_nontotal_typed_dict_missing_required():
    with pytest.raises(UnjsonifyError):
        unjsonify[ApiKey2]({"name": "key", "created_at": "2023-08-05T07:22:33"})


def test_unjsonify_nontotal_typed_dict_extra_key():
    with pytest.raises(UnjsonifyError):
        unjsonify[ApiKey2]({"value": "XXX", "treated_at": "2023-08-05T07:22:33"})


def test_unjsonify_nontotal_typed_dict_ignore_extra_key():
    with unjsonify.ignore_extra_keys():
        apikey = unjsonify[ApiKey2]({"value": "XXX", "treated_at": "2023-08-05T07:22:33"})
        assert apikey == {"value": "XXX"}


class LogEntry(NamedTuple):
    date: datetime
    message: str


entry = LogEntry(date=datetime(2023, 8, 5, 8, 32, 10), message="fail")
json = ["2023-08-05T08:32:10", "fail"]


def test_jsnoify_namedtuple():
    assert jsonify(entry) == json


def test_unjsonify_namedtuple():
    assert unjsonify[LogEntry](json) == entry


def test_unjsonify_namedtuple_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[LogEntry](["2023-08-05T08:32:10"])


class AnnotatedLogEntry(TypedDict):
    date: datetime
    message: str // Constraint.len(min=2)


def test_unjsonify_constrained():
    json = {"date": "2023-08-13T19:21:00", "message": "fail"}
    entry = AnnotatedLogEntry(date=datetime(2023, 8, 13, 19, 21, 0), message="fail")
    assert unjsonify[AnnotatedLogEntry](json) == entry

    with pytest.raises(UnjsonifyError):
        unjsonify[LogEntry]({**json, "message": "!"})


class LogEntryWithPropertyName(TypedDict):
    date: datetime // property_name("log-date")  # noqa
    message: str // Constraint.len(min=2)


def test_unjsonify_property_name_error():

    # with pytest.raises(TypeError):
    unjsonify[LogEntryWithPropertyName]


class LinkedList(TypedDict):
    value: str
    next: Optional["LinkedList"]


def test_unjsonify_recursive_type():
    json = {"value": "one", "next": {"value": "two", "next": None}}

    assert unjsonify[LinkedList](json) == json



import dataclasses


def as_field(name, arg):
    if isinstance(arg, tuple):
        # argument is a pair.
        (type_, field) = arg
        if isinstance(field, dataclasses.Field):
            # second element is a dataclass field - take it as is
            return (name, type_, field)
        else:
            # sscond element is something other - treat as a default value
            return (name, type_, dataclasses.field(default=field))
    else:
        # Argument is something that's not a tuple - interpret
        # it to be a type

        return (name, arg)


def Record(**kwargs):
    """
    Function that constructs anonoymous dataclasses.
    """

    fields = [as_field(name, arg) for (name, arg) in kwargs.items()]
    return dataclasses.make_dataclass("Record", fields)



import collections
import datetime

import pytest

from jsno.jsonify import jsonify


def test_jsonify_none():
    assert jsonify(None) is None


def test_jsonify_string():
    assert jsonify("foo") == "foo"


def test_jsonify_int():
    assert jsonify(123) == 123


def test_jsonify_tuple():
    assert jsonify(("x", 1, datetime.date(2023, 7, 15))) == ["x", 1, "2023-07-15"]


def test_jsonify_set():
    assert jsonify(set("abababa")) == ["a", "b"]


def test_jsonify_fonzenset():
    assert jsonify(frozenset("abababa")) == ["a", "b"]


def test_jsonify_dict():
    assert jsonify({"x": {"z": 123}}) == {"x": {"z": 123}}


def test_jsonify_list_of_bools():
    assert jsonify([True, False]) == [True, False]


def test_jsonify_counter():
    assert jsonify(collections.Counter("abababa") == {"a": 4, "b": 3})


def test_jsonify_dictionary_with_date_and_name():
    assert (
        jsonify({"name": "test", "date": datetime.date(2023, 7, 16)})
        ==
        {"name": "test", "date": "2023-07-16"}
    )


def test_jsonify_dictionary_with_date_keys():

    assert (
        jsonify({datetime.date(2023, 7, 16): datetime.date(2023, 7, 17)})
        ==
        {"2023-07-16": "2023-07-17"}
    )


def test_pure_json():
    item = {
        "things": [
            {
                "name": f"a{ix}",
                "scores": [1, 2, 3, ix]
            }
            for ix in range(10)
        ]
    }

    assert jsonify(item) is item


def test_jsonify_bytes():
    bs = "foobar!".encode('utf-8')
    assert jsonify(bs) == "Zm9vYmFyIQ=="


def test_jsonify_heterogenous_list():
    assert jsonify(["foo", datetime.date(2023, 7, 16)]) == ["foo", "2023-07-16"]


def test_jsonify_empty_list():
    assert jsonify([]) == []


def test_jsonify_dict_with_int_keys():
    assert jsonify({1: 2, 3: 4}) == {"1": 2, "3": 4}


def test_jsonify_error():
    with pytest.raises(TypeError):
        jsonify(lambda x: x)


def test_call_jsonify_as_type():
    assert jsonify.call_as_type(123, int) == 123
    assert jsonify.call_as_type({}, dict) == {}


def test_call_jsonify_as_type_error():
    with pytest.raises(TypeError):
        jsonify.call_as_type(123, dict)


def test_unjsonify_str_subclass():

    class SpecialString(str):
        pass

    assert jsonify(SpecialString("special")) == "special"


def test_unjsonify_float_subclass():

    class SpecialFloat(float):
        pass

    assert jsonify(SpecialFloat(123.4)) == 123.4


def test_namedtuple():
    LogEntry = collections.namedtuple("LogEntry", ["date", "message"])

    entry = LogEntry(date=datetime.date(2023, 8, 5), message="ok")
    assert jsonify(entry) == ["2023-08-05", "ok"]



import dataclasses
import functools
import threading
import time
import types

from collections.abc import Mapping
from typing import (
    Annotated, Any, Callable, Union, Literal, NewType, Self,
    get_args, get_origin, get_type_hints,
    Required, NotRequired,
)

from jsno.fields_unjsonifier import (
    UnjsonifyError, SchemaField, create_unjsonifier, typecheck, raise_error, unjsonify_context
)
from jsno.constraint import get_validators, get_class_annotations

from jsno.property_name import get_property_name
from jsno.utils import DictWithoutKey
from jsno.variant import get_variantfamily


def cast(value: Any, as_type: Any) -> Any:
    """
    Make sure that the value given has given type, and if not,
    try casting it to that type.
    """

    if isinstance(as_type, type):
        origin_type = as_type
    else:
        origin_type = get_origin(as_type)

    if isinstance(value, origin_type):
        return value

    try:
        return origin_type(value)
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


@functools.singledispatch
def unjsonify_factory(as_type):
    if dataclasses.is_dataclass(as_type):
        return get_unjsonify_dataclass(as_type)

    raise TypeError(f"Unjsonify not defined for {as_type.__qualname__}")


def contains_self_type(type_: type) -> bool:
    """
    Check if a type contains the Self type in it.
    """
    return (
        type_ is Self or
        any(contains_self_type(arg) for arg in get_args(type_))
    )


@dataclasses.dataclass(slots=True)
class ReferThrough:
    """
    Special class that defers unjsonify resolution to
    the call-time, used for handling recursive definitions.
    """
    as_type: type
    specialized: Callable | None = None

    def __call__(self, value):
        if self.specialized is None:
            self.specialized = unjsonify[self.as_type]

        return self.specialized(value)


def get_unjsonify_for_field(field_type, self_type):

    # Required check for TypedDicts
    origin = get_origin(field_type)
    if origin is NotRequired or origin is Required:
        field_type = get_args(field_type)[0]

    unjsonify_ = unjsonify[field_type]

    if not contains_self_type(field_type):
        return unjsonify_

    def wrapped(value):
        with unjsonify_context(self_type=self_type):
            return unjsonify_(value)

    return wrapped


def resolve_field_unjsonifiers(as_type, required_keys=frozenset()):
    type_hints = get_type_hints(as_type, include_extras=True)

    unjsonify._context_stack.add(as_type)
    try:
        return [
            SchemaField(
                name=name,
                json_name=json_name,
                default=Required if name in required_keys else NotRequired,
                unjsonify=get_unjsonify_for_field(type_, as_type)
            )
            for (name, type_) in type_hints.items()
            if (json_name := get_property_name(type_, name))
        ]
    finally:
        unjsonify._context_stack.remove(as_type)


def get_unjsonify_dataclass(as_type):
    if as_type in unjsonify._context_stack:
        return ReferThrough(as_type)

    unjsonifier = create_unjsonifier(
        as_type=as_type,
        fields=resolve_field_unjsonifiers(as_type),
    )

    def specialized(value):
        kwargs = unjsonifier.unjsonify_fields(value)
        try:
            return as_type(**kwargs)
        except TypeError as exc:
            detail = exc.args[0]

        raise_error(value, as_type, detail)

    return specialized


def get_unjsonify_variant(as_type, family) -> Callable:
    """
    Get the unjsonify function specialized for a variant family
    """

    # mapping from labels to corresponding unjsonifiers
    cache: dict[str, Callable] = {}

    label_name = family.label_name

    def specialized(value):
        typecheck(value, (dict, Mapping), as_type)

        # get the label property from the value
        label = value.get(label_name)
        if not isinstance(label, str):
            raise_error(value, as_type, f"missing {label}")

        unjsonify_variant = cache.get(label)
        if unjsonify_variant is None:
            variant_type = family.get_variant(label)
            if variant_type is None or not issubclass(variant_type, as_type):
                raise_error(value, as_type, f"unknown {label_name} label: {label}")

            unjsonify_variant = unjsonify.specialize(variant_type)
            cache[label] = unjsonify_variant

        # call the variant unjsonifier with the input value, but with
        # the label removed
        return unjsonify_variant(DictWithoutKey(base=value, key=label_name))

    return specialized


def get_unjsonify_literal(as_type):
    options = get_args(as_type)

    return lambda value: value if value in options else raise_error(value, as_type)


def unjsonify_self(value):
    self_type = unjsonify_context.self_type
    if self_type is None:
        raise TypeError("Self used without context")
    return unjsonify[self_type](value)


def get_validating_unjsonify(as_type, unjsonify, validators):
    if not validators:
        return unjsonify

    def specialized(value):
        result = unjsonify(value)
        for validate in validators:
            try:
                validate(result)
                continue
            except ValueError as exc:
                detail = exc.args[0]

            raise UnjsonifyError(f"Validation failed for {as_type}", detail)

        return result

    return specialized


@dataclasses.dataclass
class Unjsonify:
    def __init__(self) -> None:
        self._cache: dict[type, Callable] = {}
        self._lock: threading.RLock = threading.RLock()
        self._context_stack: set[type] = set()
        self._delay: int = 0

    def specialize(self, type_) -> Callable:
        if isinstance(type_, NewType):
            type_ = type_.__supertype__

        if type_ is Self:
            return unjsonify_self

        origin = get_origin(type_)
        if origin:
            # special cases needed for constructs in typing module, as
            # singledispatch fails cannot handle Union and Literal
            if origin is Union:
                return get_unjsonify_union(type_)
            elif origin is Literal:
                return get_unjsonify_literal(type_)

            # covers list[X], dict[K,V], etc.

        try:
            factory = unjsonify_factory.dispatch(origin or type_)
        except TypeError:
            raise UnjsonifyError(f"Cannot unjsonify as {repr(type_)} of type {type(type_)}")

        unjsonify_ = factory(type_)

        if isinstance(type_, type):
            validators = get_validators(get_class_annotations(type_))
            return get_validating_unjsonify(type_, unjsonify_, validators)

        return unjsonify_

    def _dispatch(self, type_) -> Callable:
        if isinstance(type_, type) and (family := get_variantfamily(type_)):
            return get_unjsonify_variant(type_, family)
        else:
            return self.specialize(type_)

    def __getitem__(self, type_) -> Callable:
        unjsonify = self._cache.get(type_)
        if unjsonify is not None:
            return unjsonify

        try:
            acquired = self._lock.acquire(blocking=False)
            if not acquired:
                self._lock.acquire(blocking=True)
                unjsonify = self._cache.get(type_)
                if unjsonify:
                    return unjsonify

            unjsonify = self._dispatch(type_)
            if not isinstance(unjsonify, ReferThrough):

                if self._delay:
                    time.sleep(self._delay)

                self._cache[type_] = unjsonify

            return unjsonify
        finally:
            self._lock.release()

    def register(self, type_):
        def decorator(func):
            @self.register_factory(type_)
            def _(as_type):
                return lambda value: func(value, as_type)

            return func

        return decorator

    def register_factory(self, type_):
        self._cache.clear()
        return unjsonify_factory.register(type_)

    def context(self, **kwargs):
        return unjsonify_context(**kwargs)

    def ignore_extra_keys(self):
        return self.context(on_extra_key="ignore")


unjsonify = Unjsonify()


@unjsonify.register_factory(types.UnionType)
def get_unjsonify_union(as_type):
    """
    Unjsonify a Union type. Tries each option at a time, and
    selects the first one that matches.
    """

    options = [
        unjsonify[type_option]
        for type_option in get_args(as_type)
    ]

    def specialized(value):
        for try_option in options:
            try:
                return try_option(value)
            except UnjsonifyError:
                # try next option
                continue

        raise UnjsonifyError(f"Cannot unjsonify as {as_type}", value)

    return specialized


@unjsonify.register(Any)
def _(value, as_type):
    """Unjsonify Any type: just return the value"""
    return value


@unjsonify.register_factory(Annotated)
def _(as_type):
    args = get_args(as_type)
    type_ = args[0]

    return get_validating_unjsonify(type_, unjsonify[type_], get_validators(args[1:]))



from jsno.utils import DictWithoutKey


def test_dict_without_key():
    base = {"x": 100, "y": 200}
    it = DictWithoutKey(base=base, key="x")

    assert list(it) == ["y"]
    assert len(it) == 1

    assert "y" in it
    assert "x" not in it
    assert it.get("x") is None

    assert it.get("y") == 200



import pytest

from jsno import jsonify, unjsonify, jsonify_with_method, UnjsonifyError


@jsonify_with_method
class NamedList(list):
    """
    Class for testing jsonify/unjsonify methods
    """
    def __init__(self, items=(), *, name):
        super().__init__(items)
        self.name = name

    def jsonify(self):
        return {
            "name": self.name,
            "items": jsonify[list](self)
        }

    @classmethod
    def unjsonify(cls, value):
        name = value['name']
        items = unjsonify[list](value['items'])
        return cls(items, name=name)


def test_overriding_with_jsonify_method():
    """
    Test that the jsnoify method that overrides the basic
    list jsonification works as intended.
    """

    named_list = NamedList([1, 2, 3], name="List")

    assert named_list.name == 'List'

    assert jsonify(named_list) == {'name': 'List', 'items': [1, 2, 3]}


def test_overriding_with_unjsonify_method():
    """
    Test that the unjsnoify method that overrides the basic
    list unjsonification works as intended.
    """
    assert (
        unjsonify[NamedList]({'name': 'List', 'items': [1, 2, 3]})
        ==
        NamedList([1, 2, 3], name="List")
    )


class RenamedList(NamedList):
    """
    Subclass of NmedList, to be registered with custom jsonification
    methods.
    """
    pass


@jsonify.register(RenamedList)
def _(value):
    """
    When jsonifying, add "renamed": true attribute.
    """

    # extend the NamedList's jsonification
    return {**jsonify[NamedList](value), "renamed": True}


@unjsonify.register(RenamedList)
def _(value, as_type):
    """
    When unjsonifying, check that the "renamed" flag is in the
    input data. Note: this is missing proper type checks.
    """

    if value.get("renamed") is not True:
        raise UnjsonifyError()

    return NamedList.unjsonify.__func__(as_type, value)


def test_overriding_with_jsonify_register():
    """
    Test that the registered jsonification method is called
    when jsonifying an instance of RenamedList.
    """

    named_list = RenamedList([1, 2, 3], name="List")

    assert jsonify(named_list) == {'name': 'List', 'items': [1, 2, 3], "renamed": True}


def test_overriding_with_unjsonify_register():
    """
    Test that unjsonifying gives back an instance of RenamedList, and
    make sure that the "renamed" flag is validated.
    """

    jsonified = {'name': 'List', 'items': [1, 2, 3], "renamed": True}
    unjsonified = unjsonify[RenamedList](jsonified)

    assert unjsonified == RenamedList([1, 2, 3], name="List")
    assert type(unjsonified) is RenamedList

    with pytest.raises(UnjsonifyError):
        unjsonify[RenamedList]({'name': 'List', 'items': [1, 2, 3]})



from dataclasses import dataclass

from jsno.utils import Annotation


@dataclass(frozen=True, slots=True)
class PropertyName(Annotation):
    """
    Annotation type for marking a property (a dataclass field), to
    be mapped to something other when converting to JSON.
    """

    name: str
    """The name to use in JSON"""


def property_name(name) -> PropertyName:
    return PropertyName(name)


def get_property_name(type_: type, default: str) -> str:
    """
    Get the property name for a type.

    If there is a property_name annotation attched, use it's
    name. Otherwise return the default.
    """

    annotation = PropertyName.get_annotation(type_)
    if annotation:
        return annotation.name
    else:
        return default



import dataclasses
import threading
import typing

from collections.abc import Mapping


@dataclasses.dataclass(slots=True)
class DictWithoutKey(Mapping):
    base: dict
    key: str

    def get(self, key, default=None):
        if key == self.key:
            return default
        else:
            return self.base[key]

    def __getitem__(self, key):
        if key == self.key:
            raise KeyError(key)
        return self.base[key]

    def __iter__(self):
        for it in self.base:
            if it != self.key:
                yield it

    def __len__(self):
        return len(self.base) - 1


class Context:
    def __init__(self, contextvar, values):
        self.contextvar = contextvar
        self.values = values
        self.old_values = {}

    def __enter__(self):
        for (key, val) in self.values.items():
            self.old_values[key] = getattr(self.contextvar, key)
            setattr(self.contextvar, key, val)

    def __exit__(self, type, value, tb):
        for (key, val) in self.old_values.items():
            setattr(self.contextvar, key, val)


def contextvar(**kwargs):
    defaults = kwargs

    threadlocal = threading.local()

    class Contextvar:

        def __getattr__(self, key):
            try:
                return getattr(threadlocal, key)
            except AttributeError:
                pass
            return defaults[key]

        def __setattr__(self, key, value):
            setattr(threadlocal, key, value)

        def __call__(self, **kwargs):
            return Context(self, kwargs)

    return Contextvar()


class Annotation:
    def __rfloordiv__(self, type_):
        """
        Override the // operator: Annotate a type with this annotation.

        type // Annotation(...)
        """
        return typing.Annotated[type_, self]

    @classmethod
    def get_annotation(cls, type_) -> typing.Self | None:
        """
        Get the (first) annotation of this calsss attached to a type, or
        None if it doesn't have one.
        """
        if typing.get_origin(type_) is typing.Annotated:
            for arg in typing.get_args(type_):
                if isinstance(arg, cls):
                    return arg

        return None



import dataclasses

from jsno import unjsonify, Record


def test_record():
    items = [
        {"name": "one", "age": 11},
        {"name": "two", "age": 7},
    ]

    records = unjsonify[list[Record(name=str, age=int)]](items)

    assert records[0].name == "one"
    assert records[1].age == 7


def test_record_with_default():
    items = [
        {"name": "one", "age": 11, "bool": True},
        {"name": "two", "age": 7},
    ]

    records = unjsonify[list[Record(name=str, age=int, bool=(bool, False))]](items)

    assert records[0].bool is True
    assert records[1].bool is False


def test_record_with_default_field():
    items = [
        {"name": "one", "things": [1, 2, 3]},
        {"name": "two"},
    ]

    Guy = Record(
        name=str,
        things=(list[int], dataclasses.field(default_factory=list))
    )

    records = unjsonify[list[Guy]](items)

    assert records[0].things == [1, 2, 3]
    assert records[1].things == []



"""
Tests for jsonifying and unjsonifying dataclasses and enum types.

"""

import dataclasses
import enum
import pytest

from typing import Any, Dict, List, Self

import jsno
from jsno import jsonify, unjsonify, UnjsonifyError


class Color(enum.Enum):
    Red = 1
    Green = 2
    Blue = 3


class Material(enum.IntEnum):
    Wood = 1
    Metal = 2
    Plastic = 3


def test_jsonify_enums():
    assert jsonify(Color.Blue) == "Blue"


def test_jsonify_intenums():
    assert jsonify(Material.Metal) == 2
    assert type(jsonify(Material.Metal)) is int


def test_unjsonify_enum():
    assert unjsonify[Color]("Blue") == Color.Blue


def test_unjsonify_enum_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[Color]("White")


def test_unjsonify_intenum():
    assert unjsonify[Material](2) == Material.Metal


def test_unjsonify_intenum_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[Material](0)


@dataclasses.dataclass
class Brick:
    width: int
    height: int
    color: Color
    material: Material | None = None


@dataclasses.dataclass
class Box:
    name: str
    width: float
    height: float

    bricks: list[Brick]


def test_jsonify_dataclass():
    box = Box(
        name="Testbox",
        width=1000.1,
        height=223.5,
        bricks=[
            Brick(color=Color.Green, width=6, height=1, material=Material.Plastic),
            Brick(color=Color.Red, width=4, height=1),
        ]
    )

    assert jsonify(box) == {
        "name": "Testbox",
        "width": 1000.1,
        "height": 223.5,
        "bricks": [
            {"color": "Green", "width": 6, "height": 1, "material": 3},
            {"color": "Red", "width": 4, "height": 1},
        ]
    }


def test_unjsonify_dataclass():
    json_box = {
        "name": "Testbox",
        "width": 1000.1,
        "height": 223.5,
        "bricks": [
            {"color": "Green", "width": 6, "height": 1, "material": 3},
            {"color": "Red", "width": 4, "height": 1},
        ]
    }

    box = Box(
        name="Testbox",
        width=1000.1,
        height=223.5,
        bricks=[
            Brick(color=Color.Green, width=6, height=1, material=Material.Plastic),
            Brick(color=Color.Red, width=4, height=1),
        ]
    )

    assert unjsonify[Box](json_box) == box


@dataclasses.dataclass
@jsno.jsonify_with_method
class EmailAddress:
    user: str
    domain: str

    def jsonify(self) -> str:
        return f'{self.user}@{self.domain}'

    @classmethod
    def unjsonify(cls, value):
        user, domain = value.split('@')
        return cls(user=user, domain=domain)


def test_jsonify_with_jsonify_method():
    email = EmailAddress(user="foobar", domain="example.com")

    assert jsonify(email) == "foobar@example.com"


def test_unjsonify_with_unjsonify_classmethod():
    email = unjsonify[EmailAddress]("foobar@example.com")

    assert email == EmailAddress(user="foobar", domain="example.com")


def test_ujsonify_with_missing_properties_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Brick]({"color": "Red"})


def test_ujsonify_with_not_dict_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Brick]("Something else")


def test_unjsonify_dataclass_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Brick]({})


def test_unjsonify_dataclass_extra_key_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Brick]({"color": "Red", "width": 1, "height": 1, "matrial": "xxx"})


def test_unjsonify_dataclass_extra_key_ignore():
    with unjsonify.ignore_extra_keys():
        brick = unjsonify[Brick]({"color": "Red", "width": 1, "height": 1, "matrial": "xxx"})
        assert brick == Brick(color=Color.Red, width=1, height=1)


@dataclasses.dataclass
class User:
    username: str
    password: str = 'pAssW0rd'
    metadata: List[Dict[str, Any]] = dataclasses.field(default_factory=list)


def test_jsonify_dataclass_with_default_value():
    assert (
        jsonify(User(username="usr")) ==
        {"username": "usr", "password": "pAssW0rd", "metadata": []}
    )

    assert (
        jsonify(User(username="usr", password="", metadata=[{"key": 100}])) ==
        {"username": "usr", "password": "", "metadata": [{"key": 100}]}
    )


def test_unjsonifty_dataclass_with_default_value():
    user = unjsonify[User]({"username": "usr"})

    assert user == User(username="usr")
    assert user.password == "pAssW0rd"

    user = unjsonify[User]({"username": "usr", "password": "", "metadata": [{"key": 100}]})

    assert user == User(username="usr", password="", metadata=[{"key": 100}])


@dataclasses.dataclass
@jsno.extra_data(property="metadata")
class MetaUser:
    username: str
    metadata: dict


def test_jsonify_dataclass_with_extra_data_property():
    assert (
        jsonify(MetaUser(username="usr", metadata={"tags": ["yes"]})) ==
        {"username": "usr", "tags": ["yes"]}
    )


def test_unjsonify_dataclass_with_extra_data_property():
    assert (
        unjsonify[MetaUser]({"username": "usr", "tags": ["yes"]}) ==
        MetaUser(username="usr", metadata={"tags": ["yes"]})
    )


@dataclasses.dataclass
@jsno.extra_data(ignore=True)
class Config:
    username: str


def test_unjsonify_dataclass_with_ignore_extra_data():
    assert (
        unjsonify[Config]({"username": "usr", "tags": ["yes"]}) ==
        Config(username="usr")
    )


@dataclasses.dataclass
class Thing:
    name: str
    age: int | None


def test_optional_property_without_default():
    thing = Thing(name="thing", age=None)

    json = jsonify(thing)
    assert json == {"name": "thing", "age": None}

    assert unjsonify[Thing](json) == thing


@dataclasses.dataclass
class FieldRequest:
    name: str // jsno.property_name("field-name")  # noqa
    type: str // jsno.property_name("field-type") // jsno.Constraint.len(min=1)  # noqa

    nojson: str // jsno.property_name(None) = "nope"


def test_jsonify_property_name():
    assert (
        jsonify(FieldRequest(name="NAME", type="TYPE", nojson="nonono")) ==
        {"field-name": "NAME", "field-type": "TYPE"}
    )


def test_unjsonify_property_name():
    assert (
        unjsonify[FieldRequest]({"field-name": "NAME", "field-type": "TYPE"}) ==
        FieldRequest(name="NAME", type="TYPE")
    )


def test_unjsonify_property_name_failure():
    json = {"field-name": "NAME", "field-type": "TYPE", "nojson": "no"}

    with pytest.raises(UnjsonifyError):
        unjsonify[FieldRequest](json)


@dataclasses.dataclass
class Folder:
    name: str
    subfolders: list[Self] = dataclasses.field(default_factory=list)


folder = Folder(name="main", subfolders=[Folder(name="sub")])


def test_jsonify_self_referential_dataclass():
    assert (
        jsonify(folder) ==
        {"name": "main", "subfolders": [{"name": "sub", "subfolders": []}]}
    )


def test_unjsonify_self_referential_dataclass():
    json = {"name": "main", "subfolders": [{"name": "sub", "subfolders": []}]}
    assert unjsonify[Folder](json) == folder


def test_unjsonify_self_without_context_error():
    with pytest.raises(TypeError):
        unjsonify[Self]({})


@dataclasses.dataclass
class ExtendedFolder(Folder):
    extra_flag: bool = True


def test_self_type_in_dataclass_with_inheritance():
    json = {"name": "main", "subfolders": [{"name": "sub", "subfolders": []}]}
    assert (
        unjsonify[ExtendedFolder](json) ==
        ExtendedFolder(name="main", subfolders=[ExtendedFolder(name="sub")])
    )


@dataclasses.dataclass
class Person:
    name: str
    friends: list["Friend"]


@dataclasses.dataclass
class Friend:
    person: Person
    nickname: str


def test_unjsonify_mutually_recursive_dataclasses():
    json = {
        "name": "X",
        "friends": [
            {
                "person": {"name": "Y", "friends": []},
                "nickname": "nick"
            }
        ]
    }

    assert (
        unjsonify[Person](json) ==
        Person(
            name="X",
            friends=[
                Friend(
                    person=Person(name="Y", friends=[]),
                    nickname="nick"
                )
            ]
        )
    )


class FunctionallyExtendedFoloder(Folder):
    # this is not marked with @dataclass
    pass


def test_dataclass_inheritance():
    json = {"name": "main", "subfolders": [{"name": "sub", "subfolders": []}]}
    assert (
        unjsonify[FunctionallyExtendedFoloder](json) ==
        FunctionallyExtendedFoloder(name="main", subfolders=[FunctionallyExtendedFoloder(name="sub")])
    )



import functools


@functools.singledispatch
def _get_extra_data_configuration(arg):
    return None


IgnoreExtraData = object()


def extra_data(property=None, ignore=None):
    assert property is None or ignore is None

    if ignore is True:
        property = IgnoreExtraData

    def decorator(cls):
        @_get_extra_data_configuration.register(cls)
        def _(arg):
            return property

        return cls

    return decorator


def get_extra_data_configuration(type_):
    return _get_extra_data_configuration.dispatch(type_)(type_)



from dataclasses import dataclass
from datetime import date
from random import Random

import json
import pathlib

import jsno
from jsno import jsonify, unjsonify


def test_domain_record_example():

    @dataclass
    class DomainRecord:
        domain: str
        ips: list[str]
        enabled_at: date | None = None

    domains = [
        DomainRecord(
            domain="example.com",
            ips=["93.184.216.34"],
            enabled_at=date(1992, 1, 1)
        ),
        DomainRecord(
            domain="another.example.com",
            ips=[]
        ),
    ]
    jsonified = jsno.jsonify(domains)
    assert jsonified == (
        [
            {
                "domain": "example.com",
                "ips": ["93.184.216.34"],
                "enabled_at": "1992-01-01"
            },
            {
                "domain": "another.example.com",
                "ips": [],
            }
        ]
    )

    pathlib.Path('/tmp/domains.json').write_text(json.dumps(jsonified, indent=4))

    jsonified = json.loads(pathlib.Path('/tmp/domains.json').read_text())
    domains = jsno.unjsonify[list[DomainRecord]](jsonified)

    assert domains[0].enabled_at.year == 1992


@jsno.jsonify.register(Random)
def _(value):
    # jsonify the random number generator as it's state
    return jsno.jsonify(value.getstate())


@jsno.unjsonify.register(Random)
def _(value, as_type):
    # first unjsonify the state
    state = jsno.unjsonify[tuple[int, tuple[int, ...], float | None]](value)

    # create a new Random object and install the unjsonified state
    result = Random()
    result.setstate(state)
    return result


def test_random_example():
    rng = Random()

    jsonified = jsno.jsonify(rng)

    a = rng.gauss()

    rng = jsno.unjsonify[Random](jsonified)

    b = rng.gauss()

    assert a == b


@dataclass
class APIRequest:
    class_: str // jsno.property_name("class")  # noqa
    instance_count: int // jsno.property_name("instance-count")  # noqa


def test_api_request_example():
    request = APIRequest(class_="Request", instance_count=1)
    json = jsonify(request)

    assert json == {"class": "Request", "instance-count": 1}
    assert unjsonify[APIRequest](json) == request



"""
Jsonify Python objects to JSON-compatible values, and unjsonify it back to
Python objects.

Jsonify maps any Python object to a structure consisting of dicts, lists,
and primitive values that is ready to be converted to JSON.

Unjsonify converts such JSON-structure to a given type. The type to convert
to must be given as argument.

The following types are currently supported:
    - int, float, bool
    - None (null in JSON)
    - lists and tuples
    - dicts
    - types marked as dataclass
    - enum types
    - datetime, date and time
    - complex
    - pathlib.Path
    - decimal.Decimal

Any type can be made compatible by registering custom functions for converting
to/from a json structure. Example converters for date and datetime are provided.

Note that the jsoniser unjsonisers are matched in reverse registration order,
so if there is a separate jsoniser/unjsoniser defined for both a subclass and
it's superclass, -the superclass must be registered first.

A custom class can be made compatible also by marking it with the
`jsonify_with_method` deorator, and providing jsonify method and
unjsonify class method.

"""

from jsno.constraint import Constraint, constraint
from jsno.extra_data import extra_data
from jsno.jsonify import jsonify
from jsno.jsonize import loads, dumps
from jsno.method import jsonify_with_method
from jsno.property_name import property_name
from jsno.record import Record
from jsno.schema import Schema
from jsno.standard import jsonify_as_string
from jsno.unjsonify import unjsonify, UnjsonifyError
from jsno.variant import get_variantfamily, variantfamily, variantlabel, VariantFamily

# import to register jsonifiers
import jsno.abc  # noqa
import jsno.datetime  # noqa
import jsno.tuple  # noqa


__version__ = "1.1.5"

__all__ = [
    "constraint",
    "dumps",
    "jsonify",
    "jsonify_as_string",
    "jsonify_with_method",
    "extra_data",
    "get_variantfamily",
    "loads",
    "property_name",
    "unjsonify",
    "variantfamily",
    "variantlabel",
    "Constraint",
    "Schema",
    "Record",
    "UnjsonifyError",
    "VariantFamily",
]



"""
Jsonification and unjsonification for the datetime types.

* datetime.datetime
* datetime.date
* datetime.time
* datetime.timezone

"""

import datetime


from jsno.jsonify import jsonify
from jsno.standard import jsonify_to_string
from jsno.unjsonify import unjsonify, typecheck, raise_error


# datetime.date


@jsonify.register(datetime.date)
def _(date):
    return f"{date.year}-{date.month:02}-{date.day:02}"


@unjsonify.register(datetime.date)
def _(value, as_type):
    typecheck(value, str, as_type)
    try:
        (ys, ms, ds) = value.split("-")
        return as_type(int(ys), int(ms), int(ds))
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


# datetime.time


@jsonify.register(datetime.time)
def _(value):
    return str(value)


@unjsonify.register(datetime.time)
def _(value, as_type):
    typecheck(value, str, as_type)

    try:
        return datetime.time.fromisoformat(value)
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


# datetime.datetime


def is_utc_datetime(dt) -> bool:
    return (
        dt.tzinfo is not None and
        dt.tzinfo.utcoffset(dt).total_seconds() == 0.0
    )


@jsonify.register(datetime.datetime)
def _(value):
    """
    Format the datetime as a string. Uses isoformat, except  when
    the timezone is UTC, attaches "Z" as the timezone, instead of "+00:00"
    """

    if is_utc_datetime(value):
        return f"{value.date()}T{value.time()}Z"
    else:
        return value.isoformat()


@unjsonify.register(datetime.datetime)
def _(value, as_type):
    typecheck(value, str, as_type)

    try:
        return as_type.fromisoformat(value)
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


# datetime.timedelta


jsonify.register(datetime.timedelta)(jsonify_to_string)


def make_time_component(multiplier, days, timestr, as_type=datetime.timedelta):
    if timestr.find(":") == 1:
        timestr = f"0{timestr}"

    time = datetime.time.fromisoformat(timestr)
    return as_type(
        days=multiplier * days,
        hours=multiplier * time.hour,
        minutes=multiplier * time.minute,
        seconds=multiplier * time.second,
        microseconds=multiplier * time.microsecond,
    )


@unjsonify.register(datetime.timedelta)
def _(value, as_type):
    typecheck(value, str, as_type)

    if value.startswith('-'):
        multiplier = -1
        value = value[1:]
    else:
        multiplier = 1

    parts = value.split(" days, ")

    try:
        days = 0 if len(parts) < 2 else int(parts[0])
        return make_time_component(multiplier, days, parts[-1], as_type)

    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)


# datetime.timezone


jsonify.register(datetime.timezone)(jsonify_to_string)


@unjsonify.register(datetime.timezone)
def _(value, as_type):
    typecheck(value, str, as_type)

    if value == 'UTC':
        return datetime.timezone.utc

    try:
        if value.startswith('UTC+'):
            time = make_time_component(1, 0, value[4:])
        elif value.startswith('UTC-'):
            time = make_time_component(-1, 0, value[4:])
        else:
            raise ValueError("Timezone must start with 'UTC'")

        return as_type(time)

    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)



"""
Constructing unjsonifiers that unjsonify dictionaries.
"""

import dataclasses
from typing import NotRequired, Required, get_origin

from jsno.property_name import get_property_name
from jsno.unjsonify import unjsonify, get_unjsonify_for_field
from jsno.fields_unjsonifier import ExtraKeysUnjsonifier, SchemaField


@dataclasses.dataclass
class Schema:
    schema: dict
    total: bool = True
    default_type: type | None = None
    extra_data_key: str | None = None

    def _resolve_default(self, type_, default):
        """
        Get the default value for a field.
        """

        # Required is used as the "default" default. If it's anything
        # other, just return that.
        if default is not Required:
            return default

        # check if the field type is wrapped in NotRquired or Required
        # annotations, and if so, return that
        origin = get_origin(type_)
        if origin is NotRequired or origin is Required:
            return origin

        # return Required or NotRequired, based on the `total` parameter
        # of this schema
        if self.total:
            return Required
        else:
            return NotRequired

    def _map_schema_field(self, name: str, type_, default) -> SchemaField:
        """
        Get the SchemaField for given field name, type, and default value.
        """

        return SchemaField(
            name=name,
            json_name=get_property_name(type_, name),
            default=self._resolve_default(type_, default),
            unjsonify=get_unjsonify_for_field(type_, name),
        )

    def __post_init__(self):
        fields = [
            self._map_schema_field(
                name=key,
                type_=arg[0] if isinstance(arg, tuple) else arg,
                default=arg[1] if isinstance(arg, tuple) else Required
            )
            for (key, arg) in self.schema.items()
        ]

        self._unjsonifier = ExtraKeysUnjsonifier.create(
            as_type=None,
            fields=fields,
            default_unjsonifier=self.default_type and unjsonify[self.default_type],
            extra_data_key=self.extra_data_key,
        )

    def unjsonify(self, value):
        return self._unjsonifier(value)

    def extend(self, **kwargs):
        """
        Return an extended version of the schema.

        Combines the given "schema" argument with the existing one, and replaces
        the other configuration options.
        """

        if 'schema' in kwargs:
            schema = self.schema.copy()
            for (key, val) in kwargs['schema'].items():
                if val is None:
                    schema.pop(key, None)
                else:
                    schema[key] = val
            kwargs['schema'] = schema

        return dataclasses.replace(self, **kwargs)



import collections
import datetime
import dataclasses
import os

import pytest

from jsno import jsonify, unjsonify, UnjsonifyError


def test_jsonify_mapping():
    # os.environ is a Mapping type that is not a subclass of dict

    assert jsonify(os.environ) == jsonify(dict(os.environ))


def test_jsonify_mapping_2():
    dc = jsonify(os.environ)

    assert jsonify(collections.UserDict(dc)) == dc


def test_unjsonify_mapping():
    dc = jsonify(os.environ)

    userdict = unjsonify[collections.UserDict](dc)
    assert type(userdict) is collections.UserDict
    assert userdict == dc


def test_unjsonify_untyped_dict():
    assert unjsonify[dict]({"foo": 123}) == {"foo": 123}


def test_unjsonify_bytestring_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[bytes]("foo")


def test_jsonify_mixed_set():
    assert jsonify(set((1, 2, 3, "abc", None))) == [None, 1, 2, 3, "abc"]


def test_jsonify_unsortable():

    @dataclasses.dataclass
    class Wrap:
        value: str

        def __hash__(self):
            return hash(self.value)

    jsonified = jsonify(set((Wrap("x"), Wrap("y"))))

    assert (
        sorted(jsonified, key=lambda it: it["value"]) ==
        [{"value": "x"}, {"value": "y"}]
    )


def test_typed_dict_subtype():
    # test a problematic case of subclass of a typed dict not behaving
    # as expected

    DateDict = dict[str, datetime.date]

    # check that the typed dict is converted correctly
    registry = unjsonify[DateDict]({"today": "2023-08-05"})
    assert registry == {"today": datetime.date(2023, 8, 5)}

    # define a subclass
    class DateRegistry(DateDict):
        pass

    # subclass dictionary values are NOT correctly converted
    subregistry = unjsonify[DateRegistry]({"today": "2023-08-05"})
    assert type(subregistry) == DateRegistry
    assert subregistry == {"today": "2023-08-05"}

    # attach types to circumvent problem
    DateRegistryTyped = DateRegistry[str, datetime.date]

    typedregistry = unjsonify[DateRegistryTyped]({"today": "2023-08-05"})
    assert type(typedregistry) == DateRegistry
    assert registry == {"today": datetime.date(2023, 8, 5)}



import functools

from jsno.unjsonify import unjsonify, resolve_field_unjsonifiers, ReferThrough
from jsno.fields_unjsonifier import create_unjsonifier


@functools.cache
def unjsonify_typeddict_factory(as_type):
    if as_type in unjsonify._context_stack:
        return ReferThrough(as_type)

    required_keys = as_type.__required_keys__

    return create_unjsonifier(
        as_type=as_type,
        fields=resolve_field_unjsonifiers(as_type, required_keys=required_keys),
    )



import functools
import re

from dataclasses import dataclass
from typing import Any, Callable

from jsno.utils import Annotation


@functools.cache
def get_class_annotations(class_) -> list[Annotation]:
    """
    Collect the class contraints of the base classes for a
    class into a mutable list. New constraints can be
    appended to the list, and will be there when this is
    called next time
    """
    return [
        annotation
        for base in class_.__bases__
        for annotation in get_class_annotations(base)
    ]


@functools.singledispatch
def get_annotation_validator(annotation):
    """
    Extendable type-indexed function that creates a validator
    function out of an annotation, if that's applicable.
    """
    return None


def get_validators(annotations):
    return [
        validator
        for annotation in annotations
        if (validator := get_annotation_validator(annotation))
    ]


class Constraint(Annotation):
    """
    Base class for consraints
    """

    def __new__(cls, *args, **kwargs):
        if cls is Constraint:
            # Creating a plain Constraint actually creates a
            # FunctionConstraint
            cls = FunctionConstraint

        obj = object.__new__(cls)
        obj.__init__(*args, **kwargs)
        return obj

    def __or__(self, that: "Constraint") -> "Constraint":
        """
        Combining constraints with or-operator
        """
        return OrConstraint(self, that)

    @staticmethod
    def regex(regex, name=None):
        return RegExConstraint(
            name=(name or f"Regular expression {regex}"),
            regex=re.compile(regex),
        )

    def evaluate(self, value) -> bool:
        """
        To be implemented by the subclasses
        """
        raise NotImplementedError()  # pragma: no cover

    def __call__(self, class_: type) -> type:
        """
        Using the constraint as a decorator to a dataclass
        """
        annotations = get_class_annotations(class_)
        annotations.append(self)
        return class_

    def validate(self, value):
        if not self.evaluate(value):
            if self.name:
                message = f"Violates constraint: {self.name}"
            else:
                message = "Violates a constraint"

            raise ValueError(message)


@get_annotation_validator.register(Constraint)
def _(constraint):
    return constraint.validate


@dataclass(slots=True, frozen=True)
class OrConstraint(Constraint):
    left: Constraint
    right: Constraint

    def evaluate(self, value) -> bool:
        return self.left.evaluate(value) or self.right.evaluate(value)

    @property
    def name(self):
        return f'{self.left.name} or {self.right.name}'


@dataclass(slots=True, frozen=True)
class RangeConstraint(Constraint):
    """
    Constraint for a value to be in the closed range [min..max].
    Either of the bounds may be omitted.
    """

    value_name = "Value"

    min: Any | None = None
    max: Any | None = None

    def evaluate(self, value) -> bool:
        return (
            (self.min is None or value >= self.min) and
            (self.max is None or value <= self.max)
        )

    @property
    def name(self):
        if self.min is None:
            return f"{self.value_name} must be at least {self.min}"
        elif self.max is None:
            return f"{self.value_name} must be at most {self.max}"
        else:
            return f"{self.value_name} must be in range {self.min}..{self.max}"


class LenConstraint(RangeConstraint):
    """
    Constraint for the length of a value to be in the closed range
    [min..max]. Either of the bounds may be omitted.
    """

    value_name = "Length"

    def evaluate(self, value) -> bool:
        return super().evaluate(len(value))


@dataclass(slots=True, frozen=True)
class RegExConstraint(Constraint):
    """
    Constraint on a string type, matching it with a regular expression.
    The regular expression must match the _whole_ value.
    """

    name: str
    regex: re.Pattern

    def evaluate(self, value) -> bool:
        return self.regex.fullmatch(str(value)) is not None


Constraint.range = RangeConstraint  # type: ignore
Constraint.len = LenConstraint  # type: ignore


@dataclass(slots=True, frozen=True)
class FunctionConstraint(Constraint):
    """
    Constraint that is evaluated by calling the provided boolean-valued
    function on the value.
    """
    function: Callable[[Any], bool]
    name: str | None = None

    def evaluate(self, value) -> bool:
        return self.function(value)


constraint = Constraint



import threading

from jsno.unjsonify import unjsonify


def test_concurrent_unjsonify():

    unjsonify._delay = 0.01
    try:
        def run_thread():
            assert unjsonify[tuple[str, ...]]([]) == ()

        threads = [
            threading.Thread(target=run_thread)
            for n in range(100)
        ]
        for thread in threads:
            thread.start()

        for thread in threads:
            thread.join()

    finally:
        unjsonify._delay = 0



"""
"Jsonization" functions, for turning data directly into a JSON string and
back. Mimics the standard json module's interface to provide drop-in
replacements.
"""

import json

from jsno.jsonify import jsonify
from jsno.unjsonify import unjsonify


def dumps(value, **kwargs) -> str:
    """
    Turn the argument into JSON. First jsonifies it and then calls
    the standard json.dumps with the result.
    """

    return json.dumps(jsonify(value), **kwargs)


class Loads:
    """
    Factory for type-specific loads functions
    """

    def __getitem__(self, type_):
        unjsonify_ = unjsonify[type_]
        return lambda *args, **kwargs: unjsonify_(json.loads(*args, **kwargs))


loads = Loads()



import dataclasses
import functools
import typing

from jsno.extra_data import get_extra_data_configuration
from jsno.property_name import get_property_name
from jsno.variant import get_variantfamily


"""
valid JSON types.
"""
JSON = bool | int | float | str | list["JSON"] | dict[str, "JSON"] | None


class FieldSpec(typing.NamedTuple):
    name: str
    json_name: str
    optional: bool


@dataclasses.dataclass(slots=True, frozen=True)
class DataclassJsonification:
    """
    Specialized datacalass jsonifier
    """

    label_name: str | None
    label: str | None

    extra_data_property: str | None

    fields: list[FieldSpec]

    def jsonify(self, value) -> dict[str, JSON]:
        result: dict[str, JSON] = {}

        # if the value's class is a member of a variant family,
        # first add the variant label to the jsonified result
        if self.label_name:
            result[self.label_name] = self.label

        # add regular fields
        for field in self.fields:
            val = getattr(value, field.name)
            if not (val is None and field.optional):
                result[field.json_name] = jsonify(val)

        # if extra data is defined, add it's contents
        if self.extra_data_property:
            if val := getattr(value, self.extra_data_property):
                for (key, subval) in val.items():
                    result[key] = jsonify(subval)

        return result

    @staticmethod
    def create(type_):
        family = get_variantfamily(type_)
        extra_data_property = get_extra_data_configuration(type_)

        return DataclassJsonification(
            label_name=family and family.label_name,
            label=family and family.get_label(type_),
            extra_data_property=extra_data_property,
            fields=[
                FieldSpec(field.name, json_name, field.default is None)
                for field in dataclasses.fields(type_)
                if field.name != extra_data_property
                if (json_name := get_property_name(field.type, field.name))
            ]
        )


class JsonificationCache(dict):
    """
    Specialized dictionary that creates jsonifications on demand

    Could use functools.cache, but directly using this is a
    little bit faster
    """
    def __missing__(self, key: type):
        self[key] = DataclassJsonification.create(key)
        return self[key]


jsonifications = JsonificationCache()


def jsonify_dataclass(value) -> dict[str, JSON]:
    """
    Jsonify a value whose type is a dataclass.
    """
    specialized = jsonifications[type(value)]
    return specialized.jsonify(value)


def jsonify_list(value: list) -> list[JSON]:
    """
    Jsonify a list.
    """

    count = len(value)

    if count == 0:
        # shortcut for empty lists
        return value

    ix = 1

    val_json = call_jsonify(value[0])
    if val_json is not value[0]:
        # the first value was transformed. Continue with the same
        # iterator
        result = [val_json]
    else:
        while True:
            if ix == count:
                return value

            val_json = call_jsonify(value[ix])
            if val_json is not value[ix]:
                # found the first transformed value
                result = value[:ix]
                result.append(val_json)
                ix += 1
                break
            ix += 1

    while ix < count:
        val_json = call_jsonify(value[ix])
        result.append(value[ix] if val_json is value[ix] else val_json)
        ix += 1

    return result


def jsonify_dict(value: dict) -> dict[str, JSON]:
    """
    Jsonify a dict.
    """

    result = None
    nonchange_count = 0

    for (key, val) in value.items():

        key_json = key if type(key) is str else str(jsonify(key))
        val_json = call_jsonify(val)

        if result is None:
            if key_json is key and val_json is val:
                nonchange_count += 1
                continue

            result = {}
            if nonchange_count > 0:
                for (k, v) in value.items():
                    result[k] = v
                    nonchange_count -= 1
                    if nonchange_count == 0:
                        break

        result[key_json] = val_json

    return result or value


@functools.singledispatch
def generic_jsonify(value):
    """
    Singledispatch function for defining custom jsonification methods.
    """

    if dataclasses.is_dataclass(value):
        return jsonify_dataclass(value)

    raise TypeError("Don't know how to jsonify", value, type(value))


native_types = {str, int, float, bool, type(None)}


def call_jsonify(value) -> JSON:
    """
    Call jsonify, using optimised paths for the native JSON types.
    """

    if type(value) in native_types:
        return value
    elif type(value) is list:
        return jsonify_list(value)
    elif type(value) is dict:
        return jsonify_dict(value)
    else:
        return generic_jsonify(value)


def call_jsonify_as_type(value, as_type: type) -> JSON:

    # this could be combined with call_jsonify, but it seems that
    # complicating call_jsonify has suprisingly big impact on
    # performance, so keeping the duplicated code for now

    if not issubclass(type(value), as_type):
        raise TypeError(f"Cannot jsonify {value} as {as_type}")

    if as_type in native_types:
        return value
    elif as_type is list:
        return jsonify_list(value)
    elif as_type is dict:
        return jsonify_dict(value)
    else:
        return generic_jsonify.dispatch(as_type)(value)


class Jsonify:
    """
    Singleton type for the jsonify function
    """

    def __call__(self, value) -> JSON:
        """
        Jsonify any value that supports jsonification.
        """
        return call_jsonify(value)

    def call_as_type(self, value, as_type: type) -> JSON:
        """
        Jsonify any value that supports jsonification, dispatching
        on the given type.
        """
        return call_jsonify_as_type(value, as_type)

    def __getitem__(self, type_):
        return lambda value: jsonify.call_as_type(value, type_)

    def register(self, type_):
        return generic_jsonify.register(type_)


jsonify = Jsonify()



import functools

from typing import Callable


class VariantFamily:
    """
    VariantFamily represents a variant class hierarchy.
    """

    def __init__(self, root_class: type, label_name: str):
        self.root_class = root_class
        self.label_name = label_name

        self._label_for_class: dict[type, str] = {}
        self._class_for_label: dict[str, type | None] = {}

    def get_variant(self, label: str) -> type | None:
        """
        Get the variant class corresponding to the given label,
        or None if there is no class registerd for that label.
        """

        # check the cache first, using VariantFamily as the sentinel!
        cls = self._class_for_label.get(label, VariantFamily)
        if cls is not VariantFamily:
            return cls

        # search subclasses
        cls = self._search_variant(label, self.root_class)
        self._class_for_label[label] = cls

        return cls

    def _search_variant(self, label: str, cls: type) -> type | None:

        if label == self.get_label(cls):
            return cls

        for sub in cls.__subclasses__():
            if it := self._search_variant(label, sub):
                return it

        return None

    def get_label(self, cls: type) -> str:
        """
        Get the label for a class. If not expiclitly, set, using the
        variantlabel decorator, the label is taken from the class name.
        """
        return self._label_for_class.get(cls) or cls.__name__

    def register_variant(self, cls: type, label: str):
        """
        Register a new variant to be found using the given label.
        """

        if self.get_variant(label) is not None:
            raise ValueError(f"Variant with the label '{label}' already registered")

        self._label_for_class[cls] = label

        # clear the label search cache, as there might be a change
        self._class_for_label.clear()


class OrphanVariant:
    def __init__(self, cls, label_name, label):
        self.cls = cls
        self.label_name = label_name
        self.label = label

    def get_variant(self, label: str) -> type | None:
        if label == self.label:
            return self.cls
        else:
            return None

    def get_label(self, cls: type) -> str:
        return self.label


@functools.singledispatch
def _get_variantfamily(cls: type) -> VariantFamily | OrphanVariant | None:
    """
    Get the VariantFamily instance corresponding to a class,
    or None, if it is not part of a variant hierarchy
    """
    return None


def get_variantfamily(cls: type) -> VariantFamily | OrphanVariant | None:
    """
    Get the variant family that the argument type is part of, or
    None if it is not part of any.
    """
    return _get_variantfamily.dispatch(cls)(None)


def register_variantfamily(cls, family):
    @_get_variantfamily.register(cls)
    def _(cls):
        return family


def variantfamily(label: str = "label") -> Callable[[type], type]:
    """
    Decorator for marking the root of a variant family.
    """

    def decorator(cls: type) -> type:

        # make sure that the class is not already part of a variant
        # family, e.g. through a superclass
        if (family := get_variantfamily(cls)) is not None:
            raise ValueError(
                f"Class {cls} is already a member of"
                f"a variant family {family}"
            )

        family = VariantFamily(cls, label)
        register_variantfamily(cls, family)
        return cls

    return decorator


def variantlabel(label: str | None = None, key: str | None = None) -> Callable[[type], type]:
    """
    Decorator for specifying the variant label of a class
    """

    def decorator(cls: type) -> type:
        the_label = cls.__name__ if label is None else label

        family = get_variantfamily(cls)
        if family and isinstance(family, VariantFamily):
            if key is not None and key != family.label_name:
                raise ValueError("Key must match the family label name")

            family.register_variant(cls, the_label)
        else:
            if key is None:
                raise ValueError("Key required for orphan variant")

            family = OrphanVariant(cls, key, the_label)
            register_variantfamily(cls, family)

        return cls

    return decorator



from dataclasses import dataclass

import jsno


@jsno.variantfamily(label='type')
class Expression:
    pass


@dataclass
class LiteralInt(Expression):
    value: int

    def evaluate(self, context):
        return self.value


@dataclass
class Reference(Expression):
    name: str

    def evaluate(self, context):
        return context[self.name]


@dataclass
class BinaryOperator(Expression):
    left: Expression
    right: Expression


class Add(BinaryOperator):
    def evaluate(self, context):
        return self.left.evaluate(context) + self.right.evaluate(context)


class Multiply(BinaryOperator):
    def evaluate(self, context):
        return self.left.evaluate(context) * self.right.evaluate(context)


def test_ast():
    ast = Add(LiteralInt(1), Multiply(LiteralInt(2), Reference("x")))

    assert ast.evaluate({"x": 3}) == 7

    dumped = jsno.dumps(ast, indent=4)
    assert jsno.loads[Expression](dumped) == ast



import inspect
import typing

from collections.abc import Sequence
from typing import Callable

from jsno.abc import unjsonify_sequence_factory
from jsno.unjsonify import unjsonify, typecheck, raise_error


def unjsonify_typed_factory(as_type: type) -> Callable:
    arg_types = typing.get_args(as_type)

    if arg_types and len(arg_types) == 2 and arg_types[1] is Ellipsis:
        # special case for a N-length one-type tuple (tuple[T, ...])
        # note: this accepts the empty tuple. This might not be strictly allowed
        return unjsonify_sequence_factory(as_type)

    # tuple types of the form tuple[int, str, ...] are not supported now

    unjsonifiers = [unjsonify[type_] for type_ in arg_types]

    def specialized(value):
        if len(value) != len(arg_types):
            raise_error(value, as_type)

        return as_type(
            unjsonify_(item)
            for (item, unjsonify_) in zip(value, unjsonifiers)
        )

    return specialized


@unjsonify.register_factory(tuple)
def _(as_type):
    """
    Unjsonify tuples. Handles several variants of tuples:

    * plain old untyped tuple: tuple
    * typed tuple: tuple[int, str, bool]
    * n-ary monotyped tuple: tuple[T, ...]
    * untyped namedtuple
    * typing.NamedTuple
    """

    if hasattr(as_type, '__args__'):
        return unjsonify_typed_factory(as_type)

    annotations = inspect.get_annotations(as_type)
    if annotations:
        # typing.NamedTuple

        def specialized_typed_namedtuple(value):
            typecheck(value, (list, Sequence), as_type)

            if len(annotations) != len(value):
                raise_error(value, as_type)

            return as_type(*(
                unjsonify[type_](val)
                for (val, (_, type_)) in zip(value, annotations.items())
            ))

        return specialized_typed_namedtuple

    else:
        # untyped tuple
        def specialized_untyped(value):
            if as_type is tuple:
                # just plain tuple
                return tuple(value)
            else:
                # namedtuple
                return as_type(*value)

        return specialized_untyped



from __future__ import annotations

from dataclasses import dataclass

from jsno import unjsonify


@dataclass
class Folder:
    name: str
    subfolders: list[Folder]


def test_unjsonify_recursive_type():
    json = {"name": "fldr", "subfolders": [{"name": "sub", "subfolders": []}]}
    folder = unjsonify[Folder](json)

    assert folder == Folder("fldr", [Folder("sub", [])])



from dataclasses import dataclass

import pytest

from jsno import jsonify, unjsonify, variantfamily, variantlabel, UnjsonifyError


@variantfamily(label='type')
class Expression:
    pass


@dataclass
class Literal(Expression):
    value: bool

    def evaluate(self, context):
        return self.value


@dataclass
class Variable(Expression):
    name: str

    def evaluate(self, context):
        return context[self.name]


@dataclass
@variantlabel('NOT')
class Not(Expression):
    expr: Expression

    def evaluate(self, context):
        return not self.expr.evaluate(context)


@dataclass
class BinaryExpression(Expression):
    left: Expression
    right: Expression


@dataclass
class And(BinaryExpression):
    def evaluate(self, context):
        return self.left.evaluate(context) and self.right.evaluate(context)


@dataclass
class Or(BinaryExpression):
    def evaluate(self, context):
        return self.left.evaluate(context) or self.right.evaluate(context)


expr = And(Not(Variable("x")), Or(Variable("y"), Literal(False)))


def test_expression_evaluation():
    assert expr.evaluate({"x": False, "y": True}) is True
    assert expr.evaluate({"x": False, "y": False}) is False


def test_jsonify_variant():
    jsonified = jsonify(expr)

    assert jsonified == {
        "type": "And",
        "left": {
            "type": "NOT",
            "expr": {
                "type": "Variable",
                "name": "x"
            }
        },
        "right": {
            "type": "Or",
            "left": {
                "type": "Variable",
                "name": "y"
            },
            "right": {
                "type": "Literal",
                "value": False
            }
        }
    }

    unjsonified = unjsonify[Expression](jsonified)

    assert unjsonified == expr


def test_jsonify_variant_not_subclass_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Or]({"type": "Literal", "value": False})


def test_jsonify_variant_unknown_label_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Expression]({"type": "Whatever", "value": False})


def test_jsonify_variant_no_label_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Expression]({"x-type": "Literal", "value": False})


def test_jsonify_variant_not_dict_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[Expression]("Something else")


def test_variantlabel_error():
    with pytest.raises(ValueError):
        @variantlabel('not-a-variant')
        class Foo:
            pass


def test_variantlabel_mismatch_error():
    with pytest.raises(ValueError):
        @variantlabel("Foo", key="foo")
        class FooExpression(Expression):
            pass


def test_variant_family_overlap_error():

    with pytest.raises(ValueError):
        @variantfamily(label='type')
        class DoesNowWork(Not):
            pass


def test_variant_label_overlap_error():

    with pytest.raises(ValueError):
        @variantlabel(label='NOT')
        class NotAgain(Expression):
            pass


def test_jsonify_variant_with_label_as_dict():
    with pytest.raises(UnjsonifyError):
        unjsonify[Expression]({"type": {"di": "ct"}})



import collections
import datetime
import pytest
import zoneinfo

from typing import Annotated, Any, Callable, Optional, List, Literal, NewType

from jsno.unjsonify import unjsonify, UnjsonifyError


helsinki = zoneinfo.ZoneInfo("Europe/Helsinki")
utc = zoneinfo.ZoneInfo("UTC")


def test_unjsonify_none():
    assert unjsonify[type(None)](None) is None


def test_unjsonify_any():
    assert unjsonify[Any](None) is None


def test_unjsonify_none_from_string_fails():
    with pytest.raises(UnjsonifyError):
        assert unjsonify[type(None)]("foo") is None


def test_unjsonify_string():
    assert unjsonify[str]("foo") == "foo"


def test_unjsonify_float():
    assert unjsonify[float](12.34) == 12.34


def test_unjsonify_dict():
    assert unjsonify[dict]({"x": [123]}) == {"x": [123]}


def test_unjsonify_counter():
    counter = unjsonify[collections.Counter]({"x": 12, "y": 33})
    assert isinstance(counter, collections.Counter)
    assert counter == {"x": 12, "y": 33}


def test_unjsonify_list():
    assert unjsonify[list]([1, 2, 3]) == [1, 2, 3]


def test_unjsonify_dict_with_types():
    assert unjsonify[dict[str, int]]({'x': 1, 'y': 2}) == {'x': 1, 'y': 2}


def test_unjsonify_list_of_floats():
    assert unjsonify[list[float]]([1.0, 2.0, 3.0, 0, -2]) == [1, 2, 3.0, 0, -2]


def test_unjsonify_list_of_bools():
    assert unjsonify[list[bool]]([True, False]) == [True, False]


def test_unjsonify_set_of_ints():
    assert unjsonify[set[int]]([2, 3, 4, 102]) == set((2, 3, 4, 102))


def test_unjsonify_frozenset_of_ints():
    assert unjsonify[frozenset[int]]([2, 3, 4, 102]) == frozenset((2, 3, 4, 102))


def test_unjsonify_list_of_bools_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[list[bool]]([True, 120])


def test_unjsonify_n_ary_tuple():
    assert unjsonify[tuple[float, ...]]([1.0, 2.0, 3.0]) == (1, 2, 3.0)


def test_unjsonify_untyped_tuple():
    assert unjsonify[tuple]([True, None]) == (True, None)


def test_unjsonify_untyped_empty_tuple():
    assert unjsonify[tuple]([]) == ()


def test_unjsonify_empty_tuple():
    assert unjsonify[tuple[()]]([]) == ()

    with pytest.raises(UnjsonifyError):
        unjsonify[tuple[()]]([123])


def test_ujsonify_namedtuple():
    LogEntry = collections.namedtuple("LogEntry", ["date", "message"])

    entry = LogEntry(date="2023-08-05", message="ok")
    assert unjsonify[LogEntry](["2023-08-05", "ok"]) == entry

    with pytest.raises(UnjsonifyError):
        unjsonify[tuple[LogEntry]](["2023-08-05", "ok", "more"])


def test_unjsonify_malformed_type_error():
    with pytest.raises(TypeError):
        unjsonify[tuple[bool, str, ...]]([True, "x"])


def test_unjsonify_multiple_typearg_tuple():
    assert unjsonify[tuple[float, str]]([1.0, "Yes"]) == (1, "Yes")


def test_unjsonify_multiple_typearg_tuple_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[tuple[float, str]]([1.0, "Yes", 3])


def test_unjsonify_list_of_union_types():
    assert unjsonify[List[str | int]](["Yes", 51, "No"]) == ["Yes", 51, "No"]


def test_unjsonify_list_of_union_types2():
    assert unjsonify[List[int | str]](["Yes", 51, "No"]) == ["Yes", 51, "No"]


def test_unjsonify_list_of_union_types3():
    assert unjsonify[List[list[bool] | int | str]](["Yes", 51, [True]]) == ["Yes", 51, [True]]


def test_unjsonify_optional():
    class Oneof:
        pass

    the_one = Oneof()

    @unjsonify.register(Oneof)
    def _(value, as_type):
        return the_one

    assert unjsonify[Optional[Oneof]]("whatever") is the_one


def test_unjsonify_list_of_typing_optionals():
    assert unjsonify[List[Optional[str]]](["Yes", None, "No"]) == ["Yes", None, "No"]


def test_unjsonify_list_of_option_types():
    assert unjsonify[list[str | None]](["Yes", None, "No"]) == ["Yes", None, "No"]


def test_unjsonify_literal():
    assert unjsonify[Literal["A", "B"]]("B") == "B"


def test_unjsonify_literal_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[Literal["A", "B"]]("C")


def test_unjsonify_annotated():
    assert unjsonify[Annotated[tuple[int, ...], "metadata"]]([123]) == (123,)


def test_unjsonify_dictionary_of_dates():
    assert(
        unjsonify[dict[datetime.date, datetime.date]](
            {
                "2023-07-15": "2023-07-16",
                "2023-07-16": "2023-07-17",
            }
        ) == (
            {
                datetime.date(2023, 7, 15): datetime.date(2023, 7, 16),
                datetime.date(2023, 7, 16): datetime.date(2023, 7, 17),
            }
        )
    )


def test_unjsonify_bytes():
    assert unjsonify[bytes]("Zm9vYmFyIQ==") == b"foobar!"


def test_unjsonify_str_subclass():

    class SpecialString(str):
        pass

    assert unjsonify[SpecialString]("special") == SpecialString("special")


def test_unjsonify_float_subclass():

    class SpecialFloat(float):
        pass

    assert unjsonify[SpecialFloat](6.66) == SpecialFloat(6.66)


def test_unjsonify_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[dict]("ffff")


def test_unjsonify_union_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[dict | str](None)


def test_unjsonify_not_defined():
    with pytest.raises(TypeError):
        unjsonify[Callable]("call")


def test_newtype():
    Point = NewType("Point", tuple[float, float])

    assert unjsonify[Point]([1.2, 3.4]) == (1.2, 3.4)



"""
Decorator for marking a class so that it will be jsonified by a method
call: obj.jsonify()
"""
from typing import Any

from jsno.jsonify import jsonify, JSON
from jsno.unjsonify import unjsonify


def jsonify_with_method_call(value):
    """
    Jsonify function that just calls the jsonify method of
    the value object.
    """
    return value.jsonify()


def unjsonify_with_method_call(value: JSON, as_type) -> Any:
    """
    Unjsonify fuction that just calls the unjsonify class
    method of the type.
    """
    return as_type.unjsonify(value)


def jsonify_with_method(cls: type) -> type:
    """
    Decorator for classes that registers jsonify and unjsonify
    to call the corresponding instance and class methods.
    """

    jsonify.register(cls)(jsonify_with_method_call)
    unjsonify.register(cls)(unjsonify_with_method_call)

    return cls



from dataclasses import dataclass

from jsno import jsonify, unjsonify, variantfamily


@variantfamily(label="type")
class Variant:
    ...


@dataclass
class A(Variant):
    my_int: int
    my_bool: bool = False


@dataclass
class B(Variant):
    my_int: int
    my_bool: bool = True


@dataclass
class C(Variant):
    my_str: str


@dataclass
class Container:
    objects: list[Variant]


def test_variants_container_example():

    data = {
        "objects": [
            {"type": "A", "my_int": 42},
            {"type": "C", "my_str": "hello world"},
            {"type": "B", "my_int": 123},
            {"type": "A", "my_int": 321, "my_bool": True}
        ]
    }
    # note: no auto-conversion from camelCase

    c = unjsonify[Container](data)

    assert (
        c ==
        Container(
            objects=[
                A(my_int=42, my_bool=False),
                C(my_str='hello world'),
                B(my_int=123, my_bool=True),
                A(my_int=321, my_bool=True)
            ]
        )
    )

    assert unjsonify[Container](jsonify(c)) == c



from dataclasses import dataclass
from typing import Annotated

import pytest

from jsno import unjsonify, Constraint, UnjsonifyError, constraint


def test_function_constraint():
    PositiveInt = Annotated[int, Constraint(lambda x: x > 0)]

    assert unjsonify[PositiveInt](123) == 123

    with pytest.raises(UnjsonifyError):
        unjsonify[PositiveInt](0)


def test_range():
    Number = int // Constraint.range(min=1, max=3)

    assert unjsonify[Number](1) == 1

    with pytest.raises(UnjsonifyError):
        unjsonify[Number](0)


def test_len_range():
    Identifier = str // Constraint.len(min=16, max=16)

    assert unjsonify[Identifier]("1234567890abcdef") == "1234567890abcdef"

    with pytest.raises(UnjsonifyError):
        unjsonify[Identifier]("whaat")


def test_len_min():
    Identifier = str // Constraint.len(min=16)

    assert unjsonify[Identifier]("1234567890abcdef") == "1234567890abcdef"

    with pytest.raises(UnjsonifyError):
        unjsonify[Identifier]("whaat")


def test_len_max():
    Identifier = str // Constraint.len(max=3)

    assert unjsonify[Identifier]("123") == "123"

    with pytest.raises(UnjsonifyError):
        unjsonify[Identifier]("whaat")


# defining this outside the type annotation as it confuses flake8
EmailConstraint = Constraint(lambda it: "@" in it, "Valid email address")


@dataclass
class User:
    username: str
    email: str // EmailConstraint


def test_email_contraint():
    assert (
        unjsonify[User]({"username": "usr", "email": "user@domain.com"}) ==
        User("usr", "user@domain.com")
    )


def test_regex():
    Email = str // Constraint.regex(r"[\w\.]+@([\w-]+\.)+[\w-]{2,4}")

    assert unjsonify[Email]("valid.email@example.com") == "valid.email@example.com"

    with pytest.raises(UnjsonifyError):
        unjsonify[Email]("valid.email@example@com")


def test_double_constrait():
    Aaaaa = str // Constraint.regex(r"[a]+") // Constraint.len(min=5)

    assert unjsonify[Aaaaa]("aaaaa") == "aaaaa"

    with pytest.raises(UnjsonifyError):
        unjsonify[Aaaaa]("aaaaaaah")

    with pytest.raises(UnjsonifyError):
        unjsonify[Aaaaa]("aaa")


LiteralInt = Constraint.regex("[0-9]*")
LiteralString = Constraint.regex('".*"')


@dataclass
class LiteralValue:
    value: str // (LiteralInt | LiteralString)


def test_or_constraint():
    assert(
        unjsonify[list[LiteralValue]]([{"value": "123"}, {"value": '"xxx"'}]) ==
        [LiteralValue("123"), LiteralValue('"xxx"')]
    )

    with pytest.raises(UnjsonifyError):
        unjsonify[list[LiteralValue]]([{"value": "foo"}])


@dataclass
class LiteralValues:
    values: list[str // LiteralString]


def test_embedded_constraint():
    with pytest.raises(UnjsonifyError):
        unjsonify[list[LiteralValues]]([{"values": ["foo"]}])


@dataclass
@constraint(lambda it: it.min <= it.max)
class Range:
    min: int
    max: int


def test_dataclass_constraint():
    assert unjsonify[Range]({"min": 0, "max": 99}) == Range(min=0, max=99)

    with pytest.raises(UnjsonifyError):
        unjsonify[Range]({"min": 99, "max": 0})



import copy
import dataclasses
import functools

from collections.abc import Mapping
from typing import Any, Callable, Required, NotRequired

from jsno.extra_data import get_extra_data_configuration, IgnoreExtraData
from jsno.utils import contextvar


class UnjsonifyError(TypeError):
    pass


unjsonify_context = contextvar(on_extra_key="error", self_type=None)
"""Context for passing unjsonify-time configuration down to the unjsonifiers"""


def raise_error(value: Any, as_type: Any, detail=None):
    raise UnjsonifyError(f"Cannot unjsonify as {as_type}", value, detail)


def typecheck(value: Any, jsontype: type | tuple[type, ...], as_type: Any) -> None:
    """
    Check if the value is an instance of the type given as `jsontype`
    and if not, raise an appropriate UnjsonifyError
    """

    if not isinstance(value, jsontype):
        raise_error(value, as_type)


@dataclasses.dataclass(frozen=True, slots=True)
class SchemaField:
    """
    Specification on how to map one field (either a key in a dict or
    a property in a dataclass) from JSON to the Python counterpart.
    """

    name: str
    """Name of the field in the resulting object"""

    json_name: str
    """Name of the field in JSON"""

    unjsonify: Callable
    """Function to unjsonifiy the value of the field"""

    default: Any = NotRequired
    """
    Default value for the field. Special values Required and NotRequired
    can be used instead to control if the value must be present or can
    be omitted from the result.
    """


@dataclasses.dataclass
class FieldsUnjsonifier:
    as_type: type
    fields: list[SchemaField]

    @functools.cached_property
    def source_names(self) -> frozenset[str]:
        """
        The set of keys expected to be in the source (JSON) data
        """
        return frozenset(field.json_name for field in self.fields)

    def unjsonify_fields(self, value: Mapping) -> dict:
        """
        Process a JSON object, unjsonifying each field.
        """

        typecheck(value, (dict, Mapping), self.as_type)

        result = {}
        found_count = 0

        for field in self.fields:
            if field.json_name in value:
                json_value = value[field.json_name]
                result[field.name] = field.unjsonify(json_value)
                found_count += 1
            elif field.default is Required:
                raise UnjsonifyError(f"Required key not found: {field.json_name}")
            elif field.default is NotRequired:
                pass
            else:
                result[field.name] = copy.deepcopy(field.default)

        if found_count < len(value):
            self.handle_extra_keys(value, result)

        return result

    def handle_extra_keys(self, value: Mapping, result: dict) -> None:
        if unjsonify_context.on_extra_key != "error":
            return

        extra_keys = {key for key in value if key not in result}
        if self.as_type:
            type_name = getattr(self.as_type, "__qualname__", None) or self.as_type.__name__
            for_message = f" for {type_name}: {', '.join(extra_keys)}"
        else:
            for_message = ""

        raise UnjsonifyError(f"Extra keys{for_message}: {', '.join(extra_keys)}")

    def __call__(self, value):
        return self.unjsonify_fields(value)


@dataclasses.dataclass
class ExtraKeysUnjsonifier(FieldsUnjsonifier):
    """
    Extension of FieldsUnjsonifier that handles extra keys
    in the incoming JSON data.

    The extra values are unjsonified using the default_unjsonifier.
    If extra_data_key is given, any found extra keys are inserted in
    this property. If not, the keys are inserted in the resulting
    dictionary directly.
    """

    default_unjsonifier: Callable
    """Unjsonify function to call on unexpected field values."""

    extra_data_key: str | None
    """key to the property where extra data should be put."""

    def handle_extra_keys(self, value, result):
        # choose where to put the extra values
        if self.extra_data_key is None:
            # extra_data_key not defined -> put extra data
            # directly to the result dict
            extra = result
        else:
            # extra_data_key defined -> put extra data to the property
            # pointed at by it

            if self.extra_data_key not in result:
                # if the extra data property does not exist in
                # result, create it first
                result[self.extra_data_key] = {}

            extra = result[self.extra_data_key]

        # copy the extra values
        for key in value:
            if key not in self.source_names:
                extra[key] = self.default_unjsonifier(value[key])

    @staticmethod
    def create(extra_data_key, default_unjsonifier=None, **kwargs):
        """
        Create an unjsonifier. Chooses the class depending on if
        the extra data handling arguments are present.
        """

        if default_unjsonifier is None and extra_data_key is None:
            return FieldsUnjsonifier(**kwargs)

        if extra_data_key is IgnoreExtraData:
            return IgnoreExtraKeysUnjsonifier(**kwargs)

        return ExtraKeysUnjsonifier(
            default_unjsonifier=default_unjsonifier or (lambda it: it),
            extra_data_key=extra_data_key,
            **kwargs
        )


class IgnoreExtraKeysUnjsonifier(FieldsUnjsonifier):
    """
    Extension of FieldsUnjsonifier that ignores extra keys.
    """
    def handle_extra_keys(self, value, result):
        pass


def create_unjsonifier(as_type, fields):
    """
    Create unjsonifier for a type (a dataclass or a TypedDict)
    """

    extra_data_property = get_extra_data_configuration(as_type)

    return ExtraKeysUnjsonifier.create(
        as_type=as_type,
        fields=fields,
        extra_data_key=extra_data_property,
    )



import decimal
import pathlib
import re
import types
import uuid
import zoneinfo

import pytest

from jsno import jsonify, unjsonify, UnjsonifyError, jsonify_as_string


def run_tests(value, json):
    assert jsonify(value) == json
    assert unjsonify[type(value)](json) == value


def test_jsonify_decimal():
    assert jsonify(decimal.Decimal("99.9")) == "99.9"


def test_unjsonify_decimal():
    assert unjsonify[decimal.Decimal]("99.9") == decimal.Decimal("99.9")


def test_unjsonify_decimal_from_int():
    assert unjsonify[decimal.Decimal](100) == decimal.Decimal("100")


def test_jsonify_path():
    assert jsonify(pathlib.Path("/dev/null")) == "/dev/null"


def test_unjsonify_path():
    assert unjsonify[pathlib.Path]("/dev/null") == pathlib.Path("/dev/null")


def test_jsonify_zoneinfo():
    assert (
        jsonify(zoneinfo.ZoneInfo("Europe/Helsinki")) ==
        "Europe/Helsinki"
    )


def test_unjsonify_zoneinfo():
    assert (
        unjsonify[zoneinfo.ZoneInfo]("Europe/Helsinki") ==
        zoneinfo.ZoneInfo("Europe/Helsinki")
    )


def test_unjsonify_zoneinfo_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[zoneinfo.ZoneInfo]("Europe/Vantaa")


def test_jsonify_as_string_failure():

    class ZoneInfoSub(zoneinfo.ZoneInfo):
        pass

    jsonify_as_string(ZoneInfoSub, exceptions=(TypeError,))

    with pytest.raises(zoneinfo.ZoneInfoNotFoundError):
        unjsonify[ZoneInfoSub]("Europe/Vantaa")


def test_jsonify_range():
    run_tests(range(100), {"start": 0, "stop": 100})
    run_tests(range(10, 0, -1), {"start": 10, "stop": 0, "step": -1})


def test_jsonify_range_failure():
    with pytest.raises(UnjsonifyError):
        unjsonify[range]({"start": 1, "stop": 6, "step": 0})


def test_jsonify_re_pattern():
    assert jsonify(re.compile("[a-z]+")) == "[a-z]+"


def test_unjsonify_re_pattern():
    assert unjsonify[re.Pattern]("[a-z]+") == re.compile("[a-z]+")


def test_jsonify_simplenamespace():
    namespace = types.SimpleNamespace(key="value", numbers=[1, 2, 3])

    assert jsonify(namespace) == {"key": "value", "numbers": [1, 2, 3]}


def test_unjsonify_simplenamespace():
    assert (
        unjsonify[types.SimpleNamespace]({"key": "value", "numbers": [1, 2, 3]}) ==
        types.SimpleNamespace(key="value", numbers=[1, 2, 3])
    )


def test_jsonify_uuid():
    assert(
        jsonify(uuid.UUID("12345678-1234-5678-1234-AABBCCDDEEFF")) ==
        "12345678-1234-5678-1234-aabbccddeeff"
    )


def test_unjsonify_uuid():
    assert(
        unjsonify[uuid.UUID]("12345678-1234-5678-1234-aabbccddeeff") ==
        uuid.UUID("12345678-1234-5678-1234-AABBCCDDEEFF")
    )


def test_unjsonify_uuid_error():
    with pytest.raises(UnjsonifyError):
        unjsonify[uuid.UUID]("12345678-1234-5678-1234-aabbccddxxxx")



"""
Jsonification and unjsonification for abstract base classes
"""

import base64
import typing

from collections.abc import ByteString, Mapping, Sequence, Set

from jsno.jsonify import jsonify
from jsno.typeddict import unjsonify_typeddict_factory
from jsno.unjsonify import unjsonify, typecheck, raise_error, cast


# Mapping


@jsonify.register(Mapping)
def _(value):
    return {str(jsonify(key)): jsonify(val) for (key, val) in value.items()}


@unjsonify.register_factory(Mapping)
def _(as_type):
    """
    Unjsonify any Mapping type. Expects the input value to be
    a JSON object (dict).
    """
    arg_types = typing.get_args(as_type)

    if arg_types:
        # the mapping has arguments. Expects it to have key and value
        # types like dict[str, int]
        unjsonify_key = unjsonify[arg_types[0]]
        unjsonify_val = unjsonify[arg_types[1]]

        def unjsonify_dict_with_types(value):
            typecheck(value, (dict, Mapping), as_type)

            as_dict = {
                unjsonify_key(key): unjsonify_val(val)
                for (key, val) in value.items()
            }
            return cast(as_dict, as_type)

        return unjsonify_dict_with_types

    if type(as_type) is typing._TypedDictMeta:
        # TypedDicts must be caught at this stage, as they are
        # non-istantiable subclasses of dict.
        return unjsonify_typeddict_factory(as_type)

    # monotyped case: convert from dict
    def unjonify_untyped_mapping(value):
        if not isinstance(value, (dict, Mapping)):
            raise_error(value, as_type)

        return cast(value, as_type)

    return unjonify_untyped_mapping

# Sequence


@jsonify.register(Sequence)
def jsonify_sequence(value):
    return [jsonify(val) for val in value]


@unjsonify.register_factory(Sequence)
def unjsonify_sequence_factory(as_type):
    arg_types = typing.get_args(as_type)

    if not arg_types:
        def specialized_untyped(value):
            typecheck(value, (list, Sequence), as_type)
            return cast(value, as_type)

        return specialized_untyped

    else:
        unjsonify_item = unjsonify[arg_types[0]]

        def specialized_typed(value):
            typecheck(value, (list, Sequence), as_type)
            return cast([unjsonify_item(item) for item in value], as_type)

        return specialized_typed


# Set


@jsonify.register(Set)
def _(value):
    """
    Set is not a sequence, so it needs it's own jsonifier.
    Because the order of iterating over a set is not defined, the jsonification
    tries to sort the set first, to make the results more predictable.
    """

    # if possible, sort the values first.
    try:
        value = sorted(value, key=lambda v: (type(v).__name__, v))
    except Exception:
        pass

    return jsonify_sequence(value)


unjsonify.register_factory(Set)(unjsonify_sequence_factory)


# ByteString abstract base class


@jsonify.register(ByteString)
def _(value):
    return base64.b64encode(value).decode("ascii")


@unjsonify.register(ByteString)
def _(value, as_type):
    typecheck(value, str, as_type)

    try:
        return base64.b64decode(value.encode("ascii"))
    except ValueError as exc:
        detail = exc.args[0]

    raise_error(value, as_type, detail)



from datetime import date
from typing import NotRequired

import pytest

from jsno import UnjsonifyError, property_name, Schema


schema = Schema({
    "first-name": str,
    "last-name": str,
    "dates-of-birth": list[date],
    "extra-info": (str | None, None)
})


def test_schema():
    assert (
        schema.unjsonify({
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": ["2023-08-12"]
        }) ==
        {
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [date(2023, 8, 12)],
            "extra-info": None
        }
    )


def test_schema_missing_property():
    with pytest.raises(UnjsonifyError):
        schema.unjsonify({
            "first-name": "Foo",
            "dates-of-birth": ["2023-08-12"],
        })


def test_schema_non_total():
    assert (
        schema.extend(total=False).unjsonify({
            "first-name": "Foo",
            "dates-of-birth": ["2023-08-12"],
        }) ==
        {
            "first-name": "Foo",
            "dates-of-birth": [date(2023, 8, 12)],
            "extra-info": None
        }
    )


def test_schema_extra_property():
    with pytest.raises(UnjsonifyError):
        schema.unjsonify({
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [],
            "created_at": "2023-08-13",
        })


def test_schema_default_type():
    assert(
        schema.extend(default_type=date).unjsonify({
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [],
            "created_at": "2023-08-13",
        }) ==
        {
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [],
            "extra-info": None,
            "created_at": date(2023, 8, 13)
        }
    )


def test_schema_extend_schema():
    extensions = {
        "dates-of-birth": None,
        "extra-info": None,
        "last-name": str // property_name("LASTNAME")
    }
    assert(
        schema.extend(schema=extensions, extra_data_key="extra").unjsonify({
            "first-name": "Foo",
            "LASTNAME": "Bar",
            "unknown": [123]
        }) ==
        {
            "first-name": "Foo",
            "last-name": "Bar",
            "extra": {"unknown": [123]}
        }
    )

    with pytest.raises(UnjsonifyError):
        schema.extend(schema=extensions).unjsonify({
            "first-name": "Foo",
            "LASTNAME": "Bar",
            "extra-info": "extra",
        })


def test_schema_non_required_property():
    assert(
        schema.extend(schema={"created_at": NotRequired[date]}).unjsonify({
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [],
        }) ==
        {
            "first-name": "Foo",
            "last-name": "Bar",
            "dates-of-birth": [],
            "extra-info": None,
        }
    )
